$(function() {

    var $settings = CurrentGameSettings;
    var $data = {
        // makk - 0
        // piros - 1
        // tok - 2
        // zold - 3
        // suit - Math.floor(x / 20)
        // value - x % 20
        deck: [
            7,  8,  9,  2,  3,  4,  10, 11,
            27, 28, 29, 22, 23, 24, 30, 31,
            47, 48, 49, 42, 43, 44, 50, 51,
            67, 68, 69, 62, 63, 64, 70, 71
        ],
        randomCards: [],
        messages: null
    };

    var $selectors = {
        mainscreen: '.mainscreen',
        title: '.title',
        randomCards: '.randomCards',
        buttons: '.button[data-panel]',
        buttonTrigger: '.button-title, .button-title-active',
        panels: '.panel',
        languageButtons: '.language'
    };
    var $cssClasses = {
        buttonActive: 'button-active',
        panelActive: 'panel-active',
        panelLive: 'panel-live'
    };
    var $elements = {
        mainscreen: $($selectors.mainscreen),
        title: $($selectors.title),
        randomCards: $($selectors.randomCards),
        buttons: $($selectors.buttons),
        panels: $($selectors.panels),
        languageButtons: $($selectors.languageButtons)
    };
    var $templates = {};
    $('script[type="text/template"]').each(function() { $templates[this.id] = $(this).html(); })

    var $methods = {
        translateMessages: function() {
            $('[data-i18n]').each(function() {
                $(this).html($(this).attr('data-'+$settings.language));
            });
        },
        setupBackground: function() {
            $elements.mainscreen.attr('data-texture', $settings.background);
        },
        displayLogo: function() {
            $elements.title.addClass($settings.language);
        },
        renderRandomCards: function() {
            $data.randomCards = _.shuffle($data.deck).slice(0,4);
            $elements.randomCards.html(Mustache.render($templates['template-randomcards'], $data));
        }
    };
    var $events = {
        revealStartScreen: function() {
            $methods.translateMessages();
            $methods.setupBackground();
            $methods.displayLogo();
            $methods.renderRandomCards();
        },
        languageChanged: function(button) {
            var newLanguage = button.attr('data-language');
            if (newLanguage !== $settings.language) {
                $settings.language = newLanguage;
                saveSettings($settings, function() {
                    location.reload();
                });
            }
        },
        buttonClicked: function(button) {
            var relatedPanel = button.attr('data-panel');
            if (button.hasClass($cssClasses.buttonActive)) {
                button.removeClass($cssClasses.buttonActive);
                $elements.panels.filter(relatedPanel).removeClass($cssClasses.panelActive);
                return;
            }
            $elements.buttons.removeClass($cssClasses.buttonActive);
            $elements.panels.removeClass($cssClasses.panelActive);
            button.addClass($cssClasses.buttonActive);
            $elements.panels.filter(relatedPanel).addClass($cssClasses.panelActive);
        }
    };

    $elements.buttons.on('touchstart', function(e) {
        if ($(e.target).is($selectors.buttonTrigger)) {
            $events.buttonClicked($(this));
        }
    });
    $elements.panels.on('webkitTransitionEnd transitionend', function() {
        if ($(this).hasClass($cssClasses.panelActive)) {
            $(this).addClass($cssClasses.panelLive);
        } else {
            $(this).removeClass($cssClasses.panelLive);
        }
    });
    $elements.languageButtons.on('touchstart', function() {
        $events.languageChanged($(this));
    });


    $events.revealStartScreen();

    document.addEventListener("backbutton", function() {
        if (navigator.app) {
           navigator.app.exitApp();
        }
        else if (navigator.device) {
            navigator.device.exitApp();
        }
    }, false)    

});
