$(function() {

    var $settings = CurrentGameSettings;
    var $stats = CurrentGameStats.strategia;
    var $data = {
        // makk - 0
        // piros - 1
        // tok - 2
        // zold - 3
        // suit - Math.floor(x / 20)
        // value - x % 20
        deck: [
            7,  8,  9,  2,  3,  4,  10, 11,
            27, 28, 29, 22, 23, 24, 30, 31,
            47, 48, 49, 42, 43, 44, 50, 51,
            67, 68, 69, 62, 63, 64, 70, 71
        ],
        numberOfCardsInSlot: 4,
        shrinkLimit: 10,
        shrinkMoreLimit: 16,
        timer: null,
        timerPaused: false,
        time: 0,
        timeDifference: 0,
        timePrintObject: 0,
        slots: [
            { id: 0, cards: [] },
            { id: 1, cards: [] },
            { id: 2, cards: [] },
            { id: 3, cards: [] },
            { id: 4, cards: [] },
            { id: 5, cards: [] },
            { id: 6, cards: [] },
            { id: 7, cards: [] }
        ]
    };

    var $selectors = {
        hud: '.hud',
        hudClock: '.hud-clock',
        hudHelp: '.hud-help',
        hudHelpLabel: '.hud-help .hud-label',
        table: '.table',
        talons: '.talon',
        bases: '.base',
        cards: '.card',
        cardOnTopOfDeck: '.card:not(.card__matched)',
        canvas: '#canvas',
        helpOverlay: '.overlay__help',
        winOverlay: '.overlay__win',
        loseOverlay: '.overlay__lose'
    };
    var $cssClasses = {
        cardOnTopOfDeck: 'card__topcard',
        cardMatched: 'card__matched',
        cardSelectable: 'card_selectable',
        cardSelected: 'card__selected',
        cardFlipped: 'card__faceup',
        cardInactive: 'card__inactive',
        cardAnimated: 'card__animated',
        unsolvable: 'unsolvable',
        solved: 'solved',
        clockStopped: 'hud-clock__stopped',
        clockBounce: 'hud-clock__bounce',
        overlayShown: 'overlay-visible',
        shortestGame: 'show-shortest-label',
        shrinkedPile: 'shrinked',
        shrinkedMorePile: 'shrinkedmore',
        hudHelpVisible: 'hud-help__visible'
    };
    var $elements = {
        table: $($selectors.table),
        cards: null,
        hud: $($selectors.hud),
        hudClock: $($selectors.hudClock),
        hudHelp: $($selectors.hudHelp),
        hudHelpLabel: $($selectors.hudHelpLabel),
        canvas: $($selectors.canvas),
        helpOverlay: $($selectors.helpOverlay),
        winOverlay: $($selectors.winOverlay),
        loseOverlay: $($selectors.loseOverlay)
    };
    var $templates = {};
    $('script[type="text/template"]').each(function() { $templates[this.id] = $(this).html(); })

    var $methods = {
        translateMessages: function() {
            $('[data-i18n]').each(function() {
                $(this).html($(this).attr('data-'+$settings.language));
            });
        },
        confirmQuit: function() {
            if (confirm(Mustache.render($templates['template-confirm-quit-'+$settings.language], {}))) {
                window.location = 'index.html'
            }
        },
        setupBackground: function() {
            $elements.table.attr('data-texture', $settings.background);
        },
        followsValue: function(cardToFollow, card) {
            var rank = [
                0, 7, 8, 9, 10, 2, 3, 4, 11
            ];
            return rank.indexOf(cardToFollow)+1 === rank.indexOf(card);
        },
        getCardByValue: function(value) {
            return $elements.cards.filter('[data-value="' + value + '"]').eq(0);
        },
        getCardByPosition: function(position) {
            return $elements.cards.filter('[data-slotposition="' + position + '"]').eq(0);
        },
        isCardInTopPosition: function(card) {
            var slotPosition = card.attr('data-slotposition').split('-');
            if ($data.slots[+slotPosition[0]].cards.length === +slotPosition[1]+1) {
                return true;
            }
            return false;
        },
        isCardInBase: function(card) {
            var slotPosition = card.attr('data-slotposition').split('-');
            if (+slotPosition > 3) {
                return true;
            }
            return false;
        },
        isCardSelected: function(card) {
            return card.hasClass($cssClasses.cardSelected);
        },
        isCardFlipped: function(card) {
            return card.hasClass($cssClasses.cardFlipped);
        },
        flipCard: function() {
            var done = new $.Deferred();
            var card = $methods.getCardByPosition('0-' + ($data.slots[0].cards.length-1).toString());
            if (card.hasClass($cssClasses.cardFlipped)) return;
            card.addClass($cssClasses.cardInactive).one('webkitTransitionEnd transitionend', function() {
                card.removeClass($cssClasses.cardInactive);
                done.resolve();
            }).addClass($cssClasses.cardFlipped);
            return done.promise();
        },
        markTopCards: function() {
            $elements.cards.addClass($cssClasses.cardInactive);
            $data.slots.forEach(function(slot, index) {
                var slotIndex = index;
                if (slot.cards.length) {
                    $elements.cards.filter('[data-slotposition="' + slotIndex + '-' + slot.cards[slot.cards.length-1].position + '"]').removeClass($cssClasses.cardInactive);
                }
            });
        },
        checkShrinking: function(slot) {
            var targetSlotFirstCard = $elements.cards.filter('[data-slotposition^="' + slot + '-0"]').eq(0);
            if (targetSlotFirstCard.hasClass($cssClasses.shrinkedMorePile)) { return $cssClasses.shrinkedMorePile; }
            if (targetSlotFirstCard.hasClass($cssClasses.shrinkedPile)) { return $cssClasses.shrinkedPile; }
            return '';
        },
        shrinkPiles: function() {
            var talons = '123';
            $data.slots.forEach(function(slot, index) {
                if (talons.indexOf(index) !== -1) {
                    if (slot.cards.length > $data.shrinkLimit) {
                        if (slot.cards.length > $data.shrinkMoreLimit) {
                            $elements.cards.filter('[data-slotposition^="' + index + '"]').addClass($cssClasses.shrinkedMorePile);
                        } else {
                            $elements.cards.filter('[data-slotposition^="' + index + '"]').addClass($cssClasses.shrinkedPile);
                        }
                    } else {
                        $elements.cards.filter('[data-slotposition^="' + index + '"]').removeClass($cssClasses.shrinkedPile);
                    }
                }
            });
        },
        updateNativeObjects: function(previousPosition, newPosition) {
            movedCard = $data.slots[+previousPosition[0]].cards.pop();
            movedCard.slot = +newPosition[0];
            movedCard.position = +newPosition[1];
            $data.slots[+newPosition[0]].cards.push(movedCard);
        },
        selectCard: function(card) {
            card.addClass($cssClasses.cardSelected);
        },
        deselectCard: function(card) {
            card.removeClass($cssClasses.cardSelected);
        },
        hasAlreadySelectedCard: function() {
            return $($selectors.cards).hasClass($cssClasses.cardSelected);
        },
        sameCardSelected: function(card) {
            return +$($selectors.cards).filter('.' + $cssClasses.cardSelected).attr('data-value') === +card.attr('data-value');
        },
        getSelectedCard: function() {
            return $elements.cards.filter('.' + $cssClasses.cardSelected);
        },
        cardIsInDeck: function(card) {
            var cardPosition = card.attr('data-slotposition');
            return +cardPosition[0] === 0;
        },
        cardIsInTalon: function(card) {
            var talons = '123';
            var cardPosition = card.attr('data-slotposition');
            return talons.indexOf(cardPosition[0]) !== -1;
        },
        cardIsInBase: function(card) {
            var bases = '4567';
            var cardPosition = card.attr('data-slotposition');
            return bases.indexOf(cardPosition[0]) !== -1;
        },
        hasMatch: function(card) {
            var alreadySelectedCardValue = +$('.'+$cssClasses.cardSelected).attr('data-value') % 20;
            var cardValue = +card.attr('data-value') % 20;
            return (alreadySelectedCardValue === cardValue);
        },
        moveCard: function(card, position) {
            var done = new $.Deferred();
            var previousPosition = card.attr('data-slotposition').split('-');
            var newPosition = position.split('-');
            var hackRun = false;
            var validMoves, isGameSolved;
            card.removeClass($cssClasses.cardSelected).addClass($cssClasses.cardInactive).one('webkitTransitionEnd transitionend', function() {
                if (hackRun) return;
                hackRun = true;
                $methods.updateNativeObjects(previousPosition, newPosition);
                $methods.markTopCards();
                $methods.shrinkPiles();

                validMoves = $methods.isGameSolvable();
                isGameSolved = $methods.isGameSolved();
                $methods.updateHUD(validMoves, isGameSolved);
                if (!validMoves) {
                    if (!isGameSolved) {
                        $events.gameLost();
                    } else {
                        $events.gameWon();
                    }
                }
                done.resolve();
            }).attr('data-slotposition', position).addClass($methods.checkShrinking(newPosition[0]));
            return done.promise();
        },
        isGameSolvable: function() {
            var solvable = false;
            var topCards = $data.slots.map(function(slot, index) {
                return slot.cards.length ? slot.cards[slot.cards.length-1].value % 20 : 0;
            });
            var talonCards = topCards.splice(1, 3);
            var baseCards = topCards.splice(1, 4);

            if (topCards[0] !== 0) solvable = true;
            talonCards.forEach(function(talonCard) {
                if (baseCards.some(function(baseCard) { return $methods.followsValue(baseCard, talonCard); })) {
                    solvable = true;
                }
            })

            return solvable;
        },
        isGameSolved: function() {
            var solved = false;
            var topCards = $data.slots.map(function(slot, index) {
                return slot.cards.length ? slot.cards[slot.cards.length-1].value % 20 : 0;
            });
            var talonCards = topCards.splice(1, 3);
            var baseCards = topCards.splice(1, 4);

            if (baseCards.every(function(baseCard) { return baseCard === 11; })) {
                solved = true;
            }

            return solved;
        },
        updateHUD: function(validMoves, isGameSolved) {
            if (!validMoves) {
                if (!isGameSolved) {
                    $elements.hud.addClass($cssClasses.unsolvable);
                } else {
                    $elements.hud.addClass($cssClasses.solved);
                }
            }
        },
        startClock: function() {
            $data.timer = window.setInterval(function() { $methods.updateTime(); }, 10);
        },
        pad: function(str) {
            var strC = str + '';
            var padString = '00';
            return padString.substring(0, padString.length - strC.length) + strC;
        },
        updateTime: function() {
            if ($data.timerPaused) return;
            $data.time += 10;

            $data.timeDifference = $data.time;
            $data.timePrintObject = {
                days: null,
                hours: null,
                minutes: null,
                seconds: null
            };
            $data.timePrintObject.days = $methods.pad(Math.floor($data.timeDifference / 86400000));
            $data.timeDifference -= $data.timePrintObject.days * 86400000;
            $data.timePrintObject.hours = $methods.pad(Math.floor($data.timeDifference / 3600000));
            $data.timeDifference -= $data.timePrintObject.hours * 3600000;
            $data.timePrintObject.minutes = $methods.pad(Math.floor($data.timeDifference / 60000));
            $data.timeDifference -= $data.timePrintObject.minutes * 60000;
            $data.timePrintObject.seconds = $methods.pad(Math.floor($data.timeDifference / 1000));
            $data.timeDifference -= $data.timePrintObject.seconds * 1000;

            $elements.hudClock.html(Mustache.render($templates['template-hud-clock'], $data.timePrintObject));
        },
        pauseClock: function() {
            $data.timerPaused = true;
        },
        continueClock: function() {
            $data.timerPaused = false;
        },
        stopClock: function(withBounce) {
            window.clearInterval($data.timer);
            $elements.hudClock.addClass($cssClasses.clockStopped);
            if (withBounce) $elements.hudClock.addClass($cssClasses.clockBounce);
        },
        showWinOverlay: function(shortestGame) {
            $elements.winOverlay.addClass($cssClasses.overlayShown);
            if (shortestGame) {
                $elements.winOverlay.addClass($cssClasses.shortestGame);
            }
        },
        showLoseOverlay: function() {
            $elements.loseOverlay.addClass($cssClasses.overlayShown);
        },
        showHelpOverlay: function() {
            if ($elements.hudHelp.hasClass($cssClasses.hudHelpVisible)) {
                $methods.continueClock();
                $elements.helpOverlay.removeClass($cssClasses.overlayShown);
                $elements.hudHelp.removeClass($cssClasses.hudHelpVisible);
                $elements.hudHelpLabel.html($elements.hudHelpLabel.attr('data-' + $settings.language));
            } else {
                $methods.pauseClock();
                $elements.helpOverlay.addClass($cssClasses.overlayShown);
                $elements.hudHelp.addClass($cssClasses.hudHelpVisible);
                $elements.hudHelpLabel.html($elements.hudHelpLabel.attr('data-' + $settings.language + '-opened'));
            }
        },
        showAnimation: function(shortestGame) {
            var animationOrder = [];
            var nextIndex = 0;
            var thrower;
            $data.slots.forEach(function(slot) {
                slot.cards.forEach(function(card) {
                    animationOrder.unshift($methods.getCardByValue(card.value));
                });
            });
            thrower = window.setInterval(function() {
                animationOrder[nextIndex].addClass($cssClasses.cardAnimated);
                nextIndex++;
                if (nextIndex >= $data.deck.length) {
                    window.clearInterval(thrower);
                    window.setTimeout(function() {
                        $methods.showWinOverlay(shortestGame);
                    }, 600);
                }
            }, 75);
        },
        saveStats: function(gameWon) {
            var shortestGame = ($data.time < $stats.shortestGameLength) ? true : false;
            if (gameWon) {
                $stats.totalGamesPlayed++;
                $stats.totalGamesWon++;
                $stats.gameWinRate = $stats.totalGamesWon ? Math.floor10($stats.totalGamesWon / $stats.totalGamesPlayed * 100, -2) : 0;
                $stats.totalGameLength +=  $data.time;
                $stats.averageGameLength = $stats.totalGameLength / $stats.totalGamesPlayed;
                $stats.shortestGameLength = ($data.time < $stats.shortestGameLength) ? $data.time: $stats.shortestGameLength;
            } else {
                $stats.totalGamesPlayed++;
                $stats.gameWinRate = $stats.totalGamesWon ? Math.floor10($stats.totalGamesWon / $stats.totalGamesPlayed * 100, -2) : 0;
                $stats.totalGameLength +=  $data.time;
                $stats.averageGameLength = $stats.totalGameLength / $stats.totalGamesPlayed;
            }
            saveStats('strategia', $stats);
            return shortestGame;
        }
    };
    var $events = {
        beginGame: function() {
            $methods.translateMessages();
            $methods.setupBackground();
            $events.initDeck();
            $(window).on('load', function() {
                setTimeout(function() {
                    $methods.flipCard();
                }, 500);
            });
        },

        startClock: function() {
            $methods.startClock();
        },

        gameLost: function() {
            $methods.stopClock();
            window.setTimeout(function() {
                $methods.showLoseOverlay();

                $methods.saveStats(false);
            }, 600);
        },

        gameWon: function() {
            $methods.stopClock(true);
            window.setTimeout(function() {
                var shortestGame = $methods.saveStats(true);
                $methods.showAnimation(shortestGame);
                // $methods.showWinOverlay(shortestGame);
            }, 600);
        },

        initDeck: function() {
            $data.deck = _.shuffle($data.deck);
            $data.deck.forEach(function(element, index) {
                $data.slots[0].cards.push({ slot: 0, position: index, value: element });
            });
            $elements.table.prepend(Mustache.render($templates['template-deck'], $data.slots[0]));
            $elements.cards = $($selectors.cards);
        },
        setupDeck: function() {
            var validMoves = $methods.isGameSolvable();
            var isGameSolved = $methods.isGameSolved();

            $methods.markTopCards();
            $methods.updateHUD(validMoves, isGameSolved);
            if (!validMoves) {
                if (!isGameSolved) {
                    $events.gameLost();
                } else {
                    $events.gameWon();
                }
            }
        },

        noMatchMade: function() {
            var alreadySelectedCard = $('.'+$cssClasses.cardSelected);
            alreadySelectedCard.removeClass($cssClasses.cardSelected);
        },
        matchMade: function(card) {
            var alreadySelectedCard = $('.'+$cssClasses.cardSelected);
            alreadySelectedCard.addClass($cssClasses.cardMatched).removeClass($cssClasses.cardSelected).removeClass($cssClasses.cardOnTopOfDeck);
            card.addClass($cssClasses.cardMatched).removeClass($cssClasses.cardOnTopOfDeck);

            $methods.updateNativeObjects([alreadySelectedCard, card]);

            $events.setupDeck();
        },

        deckCardSelected: function(card) {
            // is it the topmost card of deck or is it already in base?
            if (!$methods.isCardInTopPosition(card)) {
                return;
            }

            if (!$methods.isCardSelected(card) && !$methods.hasAlreadySelectedCard()) {
                $methods.selectCard(card);
                return;
            }

            if ($methods.sameCardSelected(card) || $methods.hasAlreadySelectedCard()) {
                $methods.deselectCard($methods.getSelectedCard());
                return;
            }
        },
        talonCardSelected: function(card) {
            // is it the topmost card of deck or is it already in base?
            if (!$methods.isCardInTopPosition(card)) {
                return;
            }

            if (!$methods.isCardSelected(card) && !$methods.hasAlreadySelectedCard()) {
                $methods.selectCard(card);
                return;
            }

            if ($methods.sameCardSelected(card)) {
                $methods.deselectCard(card);
                return;
            }

            if ($methods.hasAlreadySelectedCard() && !$methods.cardIsInTalon($methods.getSelectedCard())) {
                var newPosition = '';
                var newSlot = card.attr('data-slotposition').substr(0,1);
                var newPlace = +card.attr('data-slotposition').substr(2);
                newPlace++;
                newPosition = newSlot + '-' + newPlace;
                $methods.moveCard($methods.getSelectedCard(), newPosition).done(function() {
                    $methods.flipCard();
                });
            } else {
                $methods.deselectCard($methods.getSelectedCard());
            }
        },
        baseCardSelected: function(card) {
            // is it the topmost card of deck or is it already in base?
            if (!$methods.isCardInTopPosition(card)) {
                return;
            }
            if (!$methods.hasAlreadySelectedCard()) {
                return;
            }

            if ($methods.hasAlreadySelectedCard()) {
                var cardValue = +card.attr('data-value') % 20;
                var selectedCardValue = +$methods.getSelectedCard().attr('data-value') % 20;

                if ($methods.followsValue(cardValue, selectedCardValue)) {
                    var newPosition = '';
                    var newSlot = card.attr('data-slotposition').substr(0,1);
                    var newPlace = +card.attr('data-slotposition').substr(2);
                    newPlace++;
                    newPosition = newSlot + '-' + newPlace;
                    $methods.moveCard($methods.getSelectedCard(), newPosition).done(function() {
                        $methods.flipCard();
                    });
                } else {
                    $methods.deselectCard($methods.getSelectedCard());
                }
            }
            return;
        },
        talonSelected: function(talon) {
            var slotPosition = talon.attr('data-id');
            if ($methods.hasAlreadySelectedCard() && !$methods.cardIsInTalon($methods.getSelectedCard())) {
                $methods.moveCard($methods.getSelectedCard(), slotPosition + '-0').done(function() {
                    $methods.flipCard();
                });
            } else {
                $methods.deselectCard($methods.getSelectedCard());
            }
        },
        baseSelected: function(base) {
            var slotPosition = base.attr('data-id');
            if ($methods.hasAlreadySelectedCard() && $methods.getSelectedCard().attr('data-value') % 20 === 7) {
                $methods.moveCard($methods.getSelectedCard(), slotPosition + '-0').done(function() {
                    $methods.flipCard();
                });
            } else {
                $methods.deselectCard($methods.getSelectedCard());
            }
        },
        helpNeeded: function() {
            $methods.showHelpOverlay();
        }
    };


    $events.beginGame();
    
    $('body').on('touchstart', $selectors.cards, function(e) {
        if ($methods.cardIsInDeck($(this))) { $events.deckCardSelected($(this)); }
        else if ($methods.cardIsInTalon($(this))) { $events.talonCardSelected($(this)); }
        else if ($methods.cardIsInBase($(this))) { $events.baseCardSelected($(this)); }
    });
    $('body').one('touchstart', $selectors.clickableCard, function() {
        $events.startClock();
    });
    $('body').on('touchstart', $selectors.talons, function() {
        $events.talonSelected($(this));
    });
    $('body').on('touchstart', $selectors.bases, function() {
        $events.baseSelected($(this));
    });

    $elements.hudHelp.on('touchstart', function() {
        $events.helpNeeded();
    });

    document.addEventListener("backbutton", function() {
        $methods.confirmQuit();
    }, false)    
    document.addEventListener("pause", function() {
        $methods.pauseClock();
    }, false)    
    document.addEventListener("resume", function() {
        $methods.continueClock();
    }, false)    

});
