$(function() {

    var $settings = CurrentGameSettings;
    var $stats = CurrentGameStats.tiszavirag;
    var $data = {
        // makk - 0
        // piros - 1
        // tok - 2
        // zold - 3
        // suit - Math.floor(x / 20)
        // value - x % 20
        deck: [
            7,  8,  9,  2,  3,  4,  10, 11,
            27, 28, 29, 22, 23, 24, 30, 31,
            47, 48, 49, 42, 43, 44, 50, 51,
            67, 68, 69, 62, 63, 64, 70, 71
        ],
        numberOfCardsInSlot: 4,
        timer: null,
        timerPaused: false,
        time: 0,
        timeDifference: 0,
        timePrintObject: 0,
        slots: [
            { id: 0, cards: null },
            { id: 1, cards: null },
            { id: 2, cards: null },
            { id: 3, cards: null },
            { id: 4, cards: null },
            { id: 5, cards: null },
            { id: 6, cards: null },
            { id: 7, cards: null }
        ]
    };

    var $selectors = {
        hud: '.hud',
        hudClock: '.hud-clock',
        hudHelp: '.hud-help',
        hudHelpLabel: '.hud-help .hud-label',
        table: '.table',
        slots: '.slot',
        cards: '.card',
        cardOnTopOfDeck: '.card:not(.card__matched)',
        clickableCard: '.card__topcard',
        canvas: '#canvas',
        helpOverlay: '.overlay__help',
        winOverlay: '.overlay__win',
        loseOverlay: '.overlay__lose'
    };
    var $cssClasses = {
        cardOnTopOfDeck: 'card__topcard',
        cardMatched: 'card__matched',
        cardSelected: 'card__selected',
        unsolvable: 'unsolvable',
        solved: 'solved',
        clockStopped: 'hud-clock__stopped',
        clockBounce: 'hud-clock__bounce',
        overlayShown: 'overlay-visible',
        shortestGame: 'show-shortest-label',
        hudHelpVisible: 'hud-help__visible'
    };
    var $elements = {
        table: $($selectors.table),
        hud: $($selectors.hud),
        hudClock: $($selectors.hudClock),
        hudHelp: $($selectors.hudHelp),
        hudHelpLabel: $($selectors.hudHelpLabel),
        canvas: $($selectors.canvas),
        helpOverlay: $($selectors.helpOverlay),
        winOverlay: $($selectors.winOverlay),
        loseOverlay: $($selectors.loseOverlay)
    };
    var $templates = {};
    $('script[type="text/template"]').each(function() { $templates[this.id] = $(this).html(); })

    var $methods = {
        translateMessages: function() {
            $('[data-i18n]').each(function() {
                $(this).html($(this).attr('data-'+$settings.language));
            });
        },
        confirmQuit: function() {
            if (confirm(Mustache.render($templates['template-confirm-quit-'+$settings.language], {}))) {
                window.location = 'index.html'
            }
        },
        setupBackground: function() {
            $elements.table.attr('data-texture', $settings.background);
        },
        shuffleDeck: function() {
            var shuffledDeck;
            do {
                shuffledDeck = _.shuffle($data.deck);
                for (var i = 0, len = $data.slots.length; i < len; i++) {
                    $data.slots[i].cards = shuffledDeck.splice(0, $data.numberOfCardsInSlot);
                }
            } while (!$methods.isGameSolvable())
        },
        renderSlots: function() {
            $elements.table.html(Mustache.render($templates['template-slots'], $data));
        },
        markTopCards: function() {
            $($selectors.slots).each(function() {
                $(this).find($selectors.cardOnTopOfDeck).eq(0).addClass($cssClasses.cardOnTopOfDeck);
            });
        },
        updateNativeObjects: function(cards) {
            var currentCardSlot;
            for (var i = 0, len = cards.length; i < len; i++) {
                currentCardSlot = +cards[i].attr('data-slot');
                $data.slots[currentCardSlot].cards.shift();
            }
        },
        selectCard: function(card) {
            card.addClass($cssClasses.cardSelected);
        },
        deselectCard: function(card) {
            card.removeClass($cssClasses.cardSelected);
        },
        hasAlreadySelectedCard: function() {
            return $($selectors.cards).hasClass($cssClasses.cardSelected);
        },
        hasMatch: function(card) {
            var alreadySelectedCardValue = +$('.'+$cssClasses.cardSelected).attr('data-value') % 20;
            var cardValue = +card.attr('data-value') % 20;
            return (alreadySelectedCardValue === cardValue);
        },
        isGameSolvable: function() {
            var disposableCards = 0;
            var lastValues = $data.slots.map(function(currentSlot) {
                if (currentSlot.cards.length) {
                    return currentSlot.cards[0] % 20;
                }
            }).filter(function(value) {
                return value;
            });
            for (var i = 0, len = lastValues.length; i < len; i++) {
                if (lastValues.indexOf(lastValues[i]) !== lastValues.lastIndexOf(lastValues[i])) {
                    disposableCards++;
                }
            }
            return disposableCards;
        },
        isGameSolved: function() {
            var cardsOnDeck = $data.slots.map(function(currentSlot) {
                return currentSlot.cards;
            }).reduce(function(a, b) {
                return a.concat(b);
            });
            if (cardsOnDeck.length <= 2) {
                return true;
            } else {
                return false;
            }
        },
        updateHUD: function(validMoves, isGameSolved) {
            if (!validMoves) {
                if (!isGameSolved) {
                    $elements.hud.addClass($cssClasses.unsolvable);
                } else {
                    $elements.hud.addClass($cssClasses.solved);
                }
            }
        },
        startClock: function() {
            $data.timer = window.setInterval(function() { $methods.updateTime(); }, 10);
        },
        pad: function(str) {
            var strC = str + '';
            var padString = '00';
            return padString.substring(0, padString.length - strC.length) + strC;
        },
        updateTime: function() {
            if ($data.timerPaused) return;
            $data.time += 10;

            $data.timeDifference = $data.time;
            $data.timePrintObject = {
                days: null,
                hours: null,
                minutes: null,
                seconds: null
            };
            $data.timePrintObject.days = $methods.pad(Math.floor($data.timeDifference / 86400000));
            $data.timeDifference -= $data.timePrintObject.days * 86400000;
            $data.timePrintObject.hours = $methods.pad(Math.floor($data.timeDifference / 3600000));
            $data.timeDifference -= $data.timePrintObject.hours * 3600000;
            $data.timePrintObject.minutes = $methods.pad(Math.floor($data.timeDifference / 60000));
            $data.timeDifference -= $data.timePrintObject.minutes * 60000;
            $data.timePrintObject.seconds = $methods.pad(Math.floor($data.timeDifference / 1000));
            $data.timeDifference -= $data.timePrintObject.seconds * 1000;

            $elements.hudClock.html(Mustache.render($templates['template-hud-clock'], $data.timePrintObject));
        },
        pauseClock: function() {
            $data.timerPaused = true;
        },
        continueClock: function() {
            $data.timerPaused = false;
        },
        stopClock: function(withBounce) {
            window.clearInterval($data.timer);
            $elements.hudClock.addClass($cssClasses.clockStopped);
            if (withBounce) $elements.hudClock.addClass($cssClasses.clockBounce);
        },
        showWinOverlay: function(shortestGame) {
            $elements.winOverlay.addClass($cssClasses.overlayShown);
            if (shortestGame) {
                $elements.winOverlay.addClass($cssClasses.shortestGame);
            }
        },
        showLoseOverlay: function() {
            $elements.loseOverlay.addClass($cssClasses.overlayShown);
        },
        showHelpOverlay: function() {
            if ($elements.hudHelp.hasClass($cssClasses.hudHelpVisible)) {
                $methods.continueClock();
                $elements.helpOverlay.removeClass($cssClasses.overlayShown);
                $elements.hudHelp.removeClass($cssClasses.hudHelpVisible);
                $elements.hudHelpLabel.html($elements.hudHelpLabel.attr('data-' + $settings.language));
            } else {
                $methods.pauseClock();
                $elements.helpOverlay.addClass($cssClasses.overlayShown);
                $elements.hudHelp.addClass($cssClasses.hudHelpVisible);
                $elements.hudHelpLabel.html($elements.hudHelpLabel.attr('data-' + $settings.language + '-opened'));
            }
        },
        showAnimation: function(shortestGame) {
            $elements.canvas.removeClass('hidden');
            $elements.canvas = $elements.canvas.get(0);
			$elements.canvas.width = $elements.table.width();
			$elements.canvas.height = $elements.table.height();

            var topShiftValue = $elements.hud.height();
			var context = $elements.canvas.getContext( '2d' );
			var cwidth = $($selectors.cards).eq(0).width(), cwidthhalf = cwidth / 2;
			var cheight = $($selectors.cards).eq(0).height(), cheighthalf = cheight / 2;
			var cardItems = [];

			var CardItem = function (id, x, y, sx, sy) {
				if ( sx === 0 ) sx = 2;

				var cx = 0;
				var cy = 0;

				this.update = function () {
					x += sx;
					y += sy;

					if (x < ( - cwidthhalf ) || x > ( $elements.canvas.width + cwidthhalf )) {
						var index = cardItems.indexOf( this );
						cardItems.splice( index, 1 );
						return false;
					}

					if (y > $elements.canvas.height - cheighthalf) {
						y = $elements.canvas.height - cheighthalf;
						sy = - sy * 0.85;
					}

					sy += 0.98;
					context.drawImage(images[id], cx, cy, 182, 300, Math.floor( x - cwidthhalf ), Math.floor( y - cheighthalf ), cwidth, cheight);
					return true;
				}
			}

			var images = $($selectors.cards);

			var throwCard = function (id, x, y) {
				var cardItem = new CardItem(id, x, y, Math.floor( Math.random() * 12 - 6 ) * 2, - Math.random() * 32);
				cardItems.push(cardItem);
			}            

			window.setInterval(function () {
				var i = 0, l = cardItems.length;
				while ( i < l ) {
					cardItems[ i ].update() ? i ++ : l --;
				}
			}, 1000 / 24);            

            var throwerIndex = 0;
            var thrower = window.setInterval(function() {
                throwCard(throwerIndex, $elements.canvas.width/2, $elements.canvas.height/2 + topShiftValue);
                throwerIndex++;
                if (throwerIndex >= $data.deck.length) {
                    window.clearInterval(thrower);
                    window.setTimeout(function() {
                        $methods.showWinOverlay(shortestGame);
                    }, 600);
                }
            }, 75);
        },
        saveStats: function(gameWon) {
            var shortestGame = ($data.time < $stats.shortestGameLength) ? true : false;
            if (gameWon) {
                $stats.totalGamesPlayed++;
                $stats.totalGamesWon++;
                $stats.gameWinRate = $stats.totalGamesWon ? Math.floor10($stats.totalGamesWon / $stats.totalGamesPlayed * 100, -2) : 0;
                $stats.totalGameLength +=  $data.time;
                $stats.averageGameLength = $stats.totalGameLength / $stats.totalGamesPlayed;
                $stats.shortestGameLength = ($data.time < $stats.shortestGameLength) ? $data.time: $stats.shortestGameLength;
            } else {
                $stats.totalGamesPlayed++;
                $stats.gameWinRate = $stats.totalGamesWon ? Math.floor10($stats.totalGamesWon / $stats.totalGamesPlayed * 100, -2) : 0;
                $stats.totalGameLength +=  $data.time;
                $stats.averageGameLength = $stats.totalGameLength / $stats.totalGamesPlayed;
            }
            saveStats('tiszavirag', $stats);
            return shortestGame;
        }
    };
    var $events = {
        beginGame: function() {
            $methods.translateMessages();
            $methods.setupBackground();
            $methods.shuffleDeck();
            $methods.renderSlots()
            $events.setupDeck();
        },

        startClock: function() {
            $methods.startClock();
        },

        gameLost: function() {
            $methods.stopClock();
            window.setTimeout(function() {
                $methods.showLoseOverlay();

                $methods.saveStats(false);
            }, 600);
        },

        gameWon: function() {
            $methods.stopClock(true);
            window.setTimeout(function() {
                var shortestGame = $methods.saveStats(true);
                $methods.showAnimation(shortestGame);
            }, 600);
        },

        setupDeck: function() {
            var validMoves = $methods.isGameSolvable();
            var isGameSolved = $methods.isGameSolved();

            $methods.markTopCards();
            $methods.updateHUD(validMoves, isGameSolved);
            if (!validMoves) {
                if (!isGameSolved) {
                    $events.gameLost();
                } else {
                    $events.gameWon();
                }
            }
        },

        noMatchMade: function() {
            var alreadySelectedCard = $('.'+$cssClasses.cardSelected);
            alreadySelectedCard.removeClass($cssClasses.cardSelected);
        },
        matchMade: function(card) {
            var alreadySelectedCard = $('.'+$cssClasses.cardSelected);
            alreadySelectedCard.addClass($cssClasses.cardMatched).removeClass($cssClasses.cardSelected).removeClass($cssClasses.cardOnTopOfDeck);
            card.addClass($cssClasses.cardMatched).removeClass($cssClasses.cardOnTopOfDeck);

            $methods.updateNativeObjects([alreadySelectedCard, card]);

            $events.setupDeck();
        },

        cardClicked: function(card) {
            if (card.hasClass($cssClasses.cardSelected)) {
                $methods.deselectCard(card);
                return;
            }

            if (!$methods.hasAlreadySelectedCard()) {
                $methods.selectCard(card);
            } else {
                if ($methods.hasMatch(card)) {
                    $events.matchMade(card);
                } else {
                    $events.noMatchMade(card);
                }
            }
        },
        helpNeeded: function() {
            $methods.showHelpOverlay();
        }
    };


    $events.beginGame();

    $('body').on('touchstart', $selectors.clickableCard, function() {
        $events.cardClicked($(this));
    });
    $('body').one('touchstart', $selectors.clickableCard, function() {
        $events.startClock();
    });
    $elements.hudHelp.on('touchstart', function() {
        $events.helpNeeded();
    });

    document.addEventListener("backbutton", function() {
        $methods.confirmQuit();
    }, false)    
    document.addEventListener("pause", function() {
        $methods.pauseClock();
    }, false)    
    document.addEventListener("resume", function() {
        $methods.continueClock();
    }, false)    

});
